import typing as t
import urllib.parse

from flask import Flask, redirect, render_template, request, session
from flask_babel import Babel
from flask_minify import Minify

from .blog import blog
from .config import (
    Config,
    languages_dict,
)
from .db.db_loader import get_db
from .plugin_loader import plugify
from .seo import seo
from .www_handler import redirect_nonwww_to_www, redirect_www_to_nonwww


class Engine(Flask):
    def __init__(self, config: Config, db, import_name):
        super().__init__(import_name)
        self.config.from_mapping(config.model_dump(by_alias=True))
        self.db = db
        self.notifiers = []
        self.dynamic_body = ""
        self.dynamic_head = ""
        babel_translation_directories = ";".join(config.translation_directories)
        self.babel = Babel(
            self,
            locale_selector=self.get_locale,
            default_translation_directories=babel_translation_directories,
        )

    def notify(self, message: str):
        for notifier in self.notifiers:
            notifier(message)

    def add_notifier(self, notifier):
        self.notifiers.append(notifier)

    def add_dynamic_body(self, body: str):
        self.dynamic_body += body

    def add_dynamic_head(self, body: str):
        self.dynamic_head += body

    def get_locale(self) -> str:
        domain = request.headers["Host"]
        domain_to_lang = self.config.get("DOMAIN_TO_LANG")

        languages = self.config.get("LANGUAGES", {}).keys()
        backup_lang = session.get(
            "language",
            request.accept_languages.best_match(languages, "en"),
        )

        if domain_to_lang:
            lang = domain_to_lang.get(domain, backup_lang)
        else:
            lang = backup_lang

        session["language"] = lang
        return lang


def create_engine(config: Config, db) -> Engine:
    app = Engine(config, db, __name__)

    @app.before_request
    def handle_www_redirection():
        if config.use_www:
            return redirect_nonwww_to_www()
        else:
            return redirect_www_to_nonwww()

    def get_langs_domain(lang: str) -> t.Optional[str]:
        lang_cfg = config.languages.get(lang)
        if lang_cfg is None:
            return None
        return lang_cfg.domain

    @app.route("/lang/<string:lang>", methods=["GET"])
    def change_language(lang):
        if new_domain := get_langs_domain(lang):
            return redirect("http://" + new_domain, code=301)
        else:
            session["language"] = lang
            return redirect(request.referrer)

    def url_link(x: str) -> str:
        return urllib.parse.quote(x, safe="")

    @app.context_processor
    def utils():
        locale = app.get_locale()
        flag = lang.flag if (lang := config.languages.get(locale)) is not None else ""
        return {
            "app_name": config.app_name,
            "languages": languages_dict(config.languages),
            "current_flag": flag,
            "current_language": locale,
            "url_link": url_link,
            "menu_items": app.db.get_menu_items(),
            "logo_url": app.db.get_logo_url(),
            "favicon_url": app.db.get_favicon_url(),
            "font": app.db.get_font(),
            "primary_color": app.db.get_primary_color(),
            "secondary_color": app.db.get_secondary_color(),
        }

    @app.context_processor
    def dynamic_body():
        return {"dynamic_body": app.dynamic_body}

    @app.context_processor
    def dynamic_head():
        return {"dynamic_head": app.dynamic_head}

    @app.errorhandler(404)
    def page_not_found(e):
        return render_template("404.html", title="404"), 404

    return plugify(app)


def create_app_from_config(config: Config) -> Engine:
    engine = create_engine_from_config(config)
    blog_blueprint = blog.create_blog_blueprint(
        db=engine.db,
        blog_prefix=config.blog_prefix,
        locale_func=engine.get_locale,
    )
    seo_blueprint = seo.create_seo_blueprint(
        db=engine.db, config=engine.config, locale_func=engine.get_locale
    )
    engine.register_blueprint(blog_blueprint)
    engine.register_blueprint(seo_blueprint)

    Minify(app=engine, html=True, js=True, cssless=True)
    return engine


def create_engine_from_config(config: Config) -> Engine:
    """Create an engine from a config."""
    db = get_db(config.db)
    return create_engine(config, db)


def create_app(config_path: str) -> Engine:
    config = Config.parse_yaml(config_path)
    return create_app_from_config(config)



from flask import redirect
from gql import gql
from pydantic import BaseModel


def json_db_get_redirections(self):
    return self.data.get("redirections", {})


def json_file_db_get_redirections(self):
    return json_db_get_redirections(self)


def google_json_db_get_redirections(self):
    return self.data.get("redirections", {})


def graph_ql_db_get_redirections(self):
    redirections = gql(
        """
        query MyQuery{
          redirections(stage: PUBLISHED){
            source
            destination
          }
        }
        """
    )
    return {
        x["source"]: x["destination"] for x in self.client.execute(redirections)["redirections"]
    }


class Redirection(BaseModel):
    source: str
    destiny: str


def parse_redirections(config: dict[str, str]) -> list[Redirection]:
    return [Redirection(source=source, destiny=destiny) for source, destiny in config.items()]


def setup_routes(app, redirections):
    for redirection in redirections:
        func = redirect_with_name(
            redirection.destiny,
            code=301,
            name=f"{redirection.source}-{redirection.destiny}",
        )
        app.route(rule=redirection.source)(func)


def redirect_with_name(destiny, code, name):
    def named_redirect(*args, **kwargs):
        return redirect(destiny, code, *args, **kwargs)

    named_redirect.__name__ = name
    return named_redirect


def process(app, config: dict[str, str]) -> object:
    redirections = parse_redirections(config)
    setup_routes(app, redirections)
    return app



def process(app, plugin_config):
    gtm_id = plugin_config["ID"]

    head_code = (
        """<!-- Google Tag Manager -->
        <script>(function(w,d,s,l,i){w[l]=w[l]||[];w[l].push({'gtm.start':
        new Date().getTime(),event:'gtm.js'});var f=d.getElementsByTagName(s)[0],
        j=d.createElement(s),dl=l!='dataLayer'?'&l='+l:'';j.async=true;j.src=
        'https://www.googletagmanager.com/gtm.js?id='+i+dl;f.parentNode.insertBefore(j,f);
        })(window,document,'script','dataLayer','"""
        + gtm_id
        + """');</script>
        <!-- End Google Tag Manager -->
    """
    )
    app.add_dynamic_head(head_code)

    body = (
        """<!-- Google Tag Manager (noscript) -->
        <noscript><iframe src="https://www.googletagmanager.com/ns.html?id="""
        + gtm_id
        + """
        height="0" width="0" style="display:none;visibility:hidden"></iframe></noscript>
        <!-- End Google Tag Manager (noscript) -->
    """
    )
    app.add_dynamic_body(body)

    return app



import smtplib

from pydantic import BaseModel, Field


def send_mail(sender_email, password, smtp_server, port, receiver_email, subject, message):
    full_message = f"From: {sender_email}\nTo: {receiver_email}\nSubject: {subject}\n\n{message}"
    server = smtplib.SMTP_SSL(smtp_server, port)
    server.ehlo()
    server.login(sender_email, password)
    server.sendmail(sender_email, receiver_email, full_message)
    server.close()


class SendMailConfig(BaseModel):
    user: str = Field(alias="sender_email")
    password: str = Field(alias="password")
    server: str = Field(alias="smtp_server")
    port: int = Field(alias="port")
    receiver: str = Field(alias="receiver_email")
    subject: str = Field(alias="subject")


def process(app, config):
    plugin_config = SendMailConfig.model_validate(config)

    def notify(message):
        send_mail(
            sender_email=plugin_config.user,
            password=plugin_config.password,
            smtp_server=plugin_config.server,
            port=plugin_config.port,
            receiver_email=plugin_config.receiver,
            subject=plugin_config.subject,
            message=message,
        )

    app.add_notifier(notify)
    return app



import importlib.util
import os
import sys
from os.path import abspath, dirname


def find_plugin(plugin_name):
    """Find plugin by name and return it as module.
    :param plugin_name: name of plugin to find
    :return: module of plugin
    """
    plugins_dir = os.path.join(dirname(abspath(__file__)), "plugins")
    module_name = plugin_name.removesuffix(".py")
    spec = importlib.util.spec_from_file_location(
        module_name, os.path.join(plugins_dir, plugin_name, "entrypoint.py")
    )
    assert spec is not None
    plugin = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = plugin
    assert spec.loader is not None
    spec.loader.exec_module(plugin)
    return plugin


def plugify(app):
    """Load plugins and run their entrypoints.
    :param app: Flask app
    :return: Flask app
    """

    plugins_data = app.db.get_plugins_data()

    for plugin_data in plugins_data:
        plugin_config = plugin_data["config"]
        plugin_name = plugin_data["name"]
        plugin = find_plugin(plugin_name)
        plugin.process(app, plugin_config)

    return app
